import numpy as np 

def point_match(x:float, y:float, z:float, X2:np.ndarray, Y2:np.ndarray, Z2:np.ndarray, tol:float=1E-6):
    """Checks to see if x,y,z is present in a Face
        x,y,z are the point that you want to match in Face defined using X2,Y2,Z2

    Args:
        x (float): Block 1 face coordinate.  
        y (float): Block 1 face coordinate 
        z (float): Block 1 face coordinate 
        X2 (np.ndarray): X(i,j,k)
        Y2 (np.ndarray): Y(i,j,k)
        Z2 (np.ndarray): Z(i,j,k)

    Returns:
        [np.ndarray]: indicies where match is found.  (-1,-1) if no match found
    """
    location = np.zeros(X2.ndim) - 1 
    dx = x - X2
    dy = y - Y2
    dz = z - Z2
    d = np.sqrt(dx* dx + dy* dy + dz* dz)    
    val = np.amin(d)
    location = np.where(d == val)
        
    if val < tol:
        return [location[0][0], location[1][0]]
    
    return [-1,-1]
    


from typing import List, Dict, Tuple
from itertools import combinations, product, permutations
import numpy as np
from numpy.core.shape_base import block
from .block import Block, rotate_block
from .face import Face, split_face
from .connectivity import get_face_intersection
from .write import write_plot3D
from math import cos, radians, sin, sqrt, acos
from copy import deepcopy

def create_face(block:Block,imin:int,imax:int,jmin:int,jmax:int,kmin:int,kmax:int) -> Face:
    """Creates a face/surface

    Args:
        block (Block): [description]
        imin (int): [description]
        imax (int): [description]
        jmin (int): [description]
        jmax (int): [description]
        kmin (int): [description]
        kmax (int): [description]

    Returns:
        Face: [description]
    """
    f = Face(4)
    if imin==imax:
        for j in [jmin,jmax]:
            for k in [kmin,kmax]:
                f.add_vertex(block.X[imin,j,k], block.Y[imin,j,k], block.Z[imin,j,k],imin,j,k)
    if jmin==jmax:
        for i in [imin,imax]:
            for k in [kmin,kmax]:
                f.add_vertex(block.X[i,jmin,k], block.Y[i,jmin,k], block.Z[i,jmin,k],i,jmin,k)
    if kmin==kmax:
        for i in [imin,imax]:
            for j in [jmin,jmax]:
                f.add_vertex(block.X[i,j,kmin], block.Y[i,j,kmin], block.Z[i,j,kmin],i,j,kmin)
    return f


def find_periodicity(blocks:List[Block],outer_faces:List, periodic_direction:str='k', rotation_axis='x'):
    """This function is used to check for periodicity of the other faces rotated about an axis 
        The way it works is to find faces of a constant i,j, or k value

    Args:
        blocks (List[Block]): [description]
        outer_faces (List): [description]
        periodic_direction (str): either i,j,k to look for
        rotation_axis (str): either x,y,z
    """
    
    periodic_surfaces = list()      # This is the output of the code 
    periodic_angle = 0
    rotation_matrix = None

    # Check periodic within a block 
    block_match = True
    while block_match:
        block_match = False
        for indx in range(len(outer_faces)):
            block = outer_faces[indx]
            outer_faces_to_remove = list()  # Integer list of which outher surfaces to remove
            split_surfaces = list()         # List of split but free surfaces, this will be appended to outer_faces_to_remove list

            blk_indx = block['index']
            surfaces = block['surfaces']   # outer surfaces for a given block 
            surface_combos = list(combinations(range(len(surfaces)),2))
            for s1,s2 in surface_combos:
                surf1_index = s1
                surf2_index = s2            
                # Check if surfaces are periodic with each other
                surface1 = create_face(blocks[blk_indx], surfaces[s1]['IMIN'], surfaces[s1]['IMAX'], 
                                                        surfaces[s1]['JMIN'], surfaces[s1]['JMAX'], 
                                                        surfaces[s1]['KMIN'], surfaces[s1]['KMAX'])

                surface2 = create_face(blocks[blk_indx], surfaces[s2]['IMIN'], surfaces[s2]['IMAX'], 
                                                surfaces[s2]['JMIN'], surfaces[s2]['JMAX'], 
                                                surfaces[s2]['KMIN'], surfaces[s2]['KMAX'])
                if periodic_direction.lower() == "i":
                    if (surfaces[s1]['IMIN'] == surfaces[s1]['IMAX']) and (surfaces[s2]['IMIN'] == surfaces[s2]['IMAX']): 
                        periodic, split_surf, faces_to_remove, angle, rot_mat = __periodicity_within_block__(surface1,surface2,blocks,surf1_index, surf2_index, blk_indx)
                        if periodic:
                            periodic_surfaces.append(periodic)
                            outer_faces_to_remove.extend(faces_to_remove)
                            split_surfaces.extend(split_surf)
                            periodic_angle = angle
                            rotation_matrix = rot_mat
                            block_match = True
                        break

                if periodic_direction.lower() == "i":
                    if (surfaces[s1]['IMIN'] == surfaces[s1]['IMAX']) and (surfaces[s2]['IMIN'] == surfaces[s2]['IMAX']): 
                        periodic, split_surf,faces_to_remove, angle, rot_mat = __periodicity_within_block__(surface1,surface2,blocks,surf1_index, surf2_index, blk_indx)
                        if periodic:
                            periodic_surfaces.append(periodic)
                            outer_faces_to_remove.extend(faces_to_remove)
                            split_surfaces.extend(split_surf)
                            periodic_angle = angle
                            rotation_matrix = rot_mat
                            block_match = True
                        break

                if periodic_direction.lower() == 'k':       # constant k between surfaces 
                    if (surfaces[s1]['KMIN'] == surfaces[s1]['KMAX']) and (surfaces[s2]['KMIN'] == surfaces[s2]['KMAX']): 
                        periodic, split_surf, faces_to_remove, angle, rot_mat = __periodicity_within_block__(surface1,surface2,blocks,surf1_index, surf2_index, blk_indx)
                        if periodic:
                            periodic_surfaces.append(periodic)
                            outer_faces_to_remove.extend(faces_to_remove)
                            split_surfaces.extend(split_surf)
                            periodic_angle = angle
                            rotation_matrix = rot_mat
                            block_match = True
                        break
                    
            if block_match:
                break
        # House keeping - Remove outer surfaces and add in the split surfaces 
        block['surfaces'] = [block['surfaces'][o] for o in range(len(block['surfaces'])) if o not in outer_faces_to_remove]
        outer_faces[indx]['surfaces'] = block['surfaces']
        for s in split_surfaces:
            outer_faces[indx]['surfaces'].append({'IMIN':s.IMIN,'JMIN':s.JMIN,'KMIN':s.KMIN,'IMAX':s.IMAX,'JMAX':s.JMAX,'KMAX':s.KMAX,'id':0})
    
    block_outer_faces_to_remove = list()
    if rotation_matrix is not None:
        # Check periodic outer_faces from block to block 
        block_combos = list(combinations(range(len(outer_faces)),2))
        for b1,b2 in block_combos: 
            # Get outer surfaces for block 1 and block 2
            # https://www.geeksforgeeks.org/python-program-to-get-all-unique-combinations-of-two-lists/ 
            for b1_surf_indx in range(len(outer_faces[b1]['surfaces'])):
                for b2_surf_indx in range(len(outer_faces[b2]['surfaces'])):
                    b1_surf = outer_faces[b1]['surfaces'][b1_surf_indx]
                    b2_surf = outer_faces[b2]['surfaces'][b2_surf_indx]

                    if (b1_surf['KMIN'] == b1_surf['KMAX']) and (b2_surf['KMIN'] == b2_surf['KMAX']):
                        surface1 = create_face(blocks[b1], b1_surf['IMIN'], b1_surf['IMAX'], 
                                                            b1_surf['JMIN'], b1_surf['JMAX'], 
                                                            b1_surf['KMIN'], b1_surf['KMAX'])

                        surface2 = create_face(blocks[b2], b2_surf['IMIN'], b2_surf['IMAX'], 
                                                            b2_surf['JMIN'], b2_surf['JMAX'], 
                                                            b2_surf['KMIN'], b2_surf['KMAX'])

                        # Step 1: Try rotating surface 1 and check for intersections 
                        block1_rotated = rotate_block(blocks[b1], rotation_matrix)
                        # write_plot3D('test_rotated.xyz',[block1_rotated],binary=True) # debug purposes
                        df, split_face1, split_face2 = get_face_intersection(surface1,surface2,block1_rotated,blocks[b2],tol=1E-5)
                        if len(df)>0:                          # Try with first block rotated
                            if periodic_direction == "i":
                                df['j1'] += surface1.JMIN
                                df['k1'] += surface1.KMIN
                                df['j2'] += surface2.JMIN
                                df['k2'] += surface2.KMIN

                            if periodic_direction == "j":
                                df['i1'] += surface1.IMIN
                                df['k1'] += surface1.KMIN
                                df['i2'] += surface2.IMIN
                                df['k2'] += surface2.KMIN

                            if periodic_direction == "k":
                                df['i1'] += surface1.IMIN
                                df['j1'] += surface1.JMIN
                                df['i2'] += surface2.IMIN
                                df['j2'] += surface2.JMIN
                        elif len(df)==0:                        # Try with second block rotated
                            # Step 2: if Step 1 does not yield a match then rotate surface 2 and check for intersections 
                            block2_rotated = rotate_block(blocks[b2], rotation_matrix)
                            df, split_face1, split_face2 = get_face_intersection(surface1,surface2,blocks[b1],block2_rotated,tol=1E-5)
                            if periodic_direction == "i":
                                df['j1'] += surface1.JMIN
                                df['k1'] += surface1.KMIN
                                df['j2'] += surface2.JMIN
                                df['k2'] += surface2.KMIN

                            if periodic_direction == "j":
                                df['i1'] += surface1.IMIN
                                df['k1'] += surface1.KMIN
                                df['i2'] += surface2.IMIN
                                df['k2'] += surface2.KMIN

                            if periodic_direction == "k":
                                df['i1'] += surface1.IMIN
                                df['j1'] += surface1.JMIN
                                df['i2'] += surface2.IMIN
                                df['j2'] += surface2.JMIN
                            
                        if len(df)>0:
                            # Here we save the periodicity 
                            periodic_surfaces.append({
                                                        'block1':{
                                                            'index':b1,'IMIN':df.iloc[0]['i1'],'JMIN':df.iloc[0]['j1'],'KMIN':df.iloc[0]['k1'],
                                                            'IMAX':df.iloc[-1]['i1'],'JMAX':df.iloc[-1]['j1'],'KMAX':df.iloc[-1]['k1']
                                                            },
                                                        'block2':{
                                                                'index':b2,'IMIN':df.iloc[0]['i2'],'JMIN':df.iloc[0]['j2'],'KMIN':df.iloc[0]['k2'],
                                                                'IMAX':df.iloc[-1]['i2'],'JMAX':df.iloc[-1]['j2'],'KMAX':df.iloc[-1]['k2']
                                                            },
                                                        })
                            # query faces to remove
                            block_outer_faces_to_remove.append({'block_indx':b1,'surface_indx':b1_surf_indx})
                            block_outer_faces_to_remove.append({'block_indx':b2,'surface_indx':b2_surf_indx})
        
        # Lets remove the outer faces
        outer_faces_to_keep = list()
        for b in range(len(outer_faces)):
            outer_faces_to_keep.append({'index':b, 'surfaces':[]})
            for s in range(len(outer_faces[b]['surfaces'])):                
                lets_remove = [True for to_remove in block_outer_faces_to_remove if to_remove['block_indx'] == b and to_remove['surface_indx'] == s] # Return true if we need to remove
                if not lets_remove: # if lets_remove is false then we want to run this statement to keep 
                    outer_faces_to_keep[b]['surfaces'].append(outer_faces[b]['surfaces'][s])
    return periodic_surfaces, outer_faces_to_keep
                        


def linear_real_transform(face1:Face,face2:Face) -> Tuple:
    """Computes the rotation angle from Face1 to Face2. This can be used to check if the faces are periodic 
        This function assumes the rotation axis is in the "x" direction 

    Reference:
        - Linear Real Transforms from GlennHT https://gitlab.grc.nasa.gov/lte-turbo/GlennHT/-/blob/master/src/M_ccMBMesh.F See computeLRT
        
    Args:
        Face1 (Face): Face to rotate
        Face2 (Face): Face to rotate to

    Returns:
        (tuple): tuple containing:

            - **ang** (float): rotation angle
            - **rotation_matrix** (numpy.ndarray): Rotation matrix 3x3
        
    """

    cTo3,cTo1 = face1.get_corners() 
    cFrom3,cFrom1 = face2.get_corners()

    dTo  = np.array(cTo3).transpose() - np.array(cTo1).transpose()                      # difference in corner points = diagonal vector for Face 1
    ldTo=np.sqrt(np.sum(dTo*dTo))
    if ldTo > 0:
        dTo=dTo/ldTo
    
    dFrom = np.array(cFrom3).transpose() - np.array(cFrom1).transpose()                 # difference in corner points = diagonal vector for Face 2
    ldFrom = np.sqrt(np.sum(dFrom*dFrom))
    if( ldFrom > 0 ):
        dFrom=dFrom/ldFrom
    
    dotprod = np.sum(dTo * dFrom)
    
    if( abs(dotprod-1) < 1E-10 ): # Case of no rotation
        ang = 0

        rotation_matrix  = np.zeros(shape=(3,3))
    else:
        #Compute the angle of rotation  
        cosAng=(dTo[1]*dFrom[1]+dTo[2]*dFrom[2])/sqrt(dTo[1]*dTo[1]+dTo[2]*dTo[2])/sqrt(dFrom[1]*dFrom[1]+dFrom[2]*dFrom[2])
        sinAng=(dTo[2]*dFrom[1]-dTo[1]*dFrom[2])/sqrt(dTo[1]*dTo[1]+dTo[2]*dTo[2])/sqrt(dFrom[1]*dFrom[1]+dFrom[2]*dFrom[2])
        ang=acos(cosAng)

        rotation_matrix = [ [1, 0, 0],
                            [0, cosAng, -sinAng],
                            [0, sinAng, cosAng] ]
        if( sinAng < 0 ):
            ang*=-1
    return ang, rotation_matrix

def __periodicity_within_block__(surface1:Face, surface2:Face,blocks:List[Block],surface1_indx:int, surface2_indx:int, block_indx:int):
    """General function to find periodicity within a given block. 

    Steps:
        - 1: Take the surface with the shorter diagonal. Rotate the face about the x-axis. We call this surface-1-rotated. 
        - 2: Shorten surface 2 to match the I and J bounds of surface 1.
        - 3: Do a Linear Real Transform (LRT) Check 

    Args:
        surface1 (Face): [description]
        surface2 (Face): [description]
        blocks (List[Block]): List of blocks
        surface1_indx (int): [description]
        surface2_indx (int): [description]

    Returns:
        [type]: [description]
    """
    periodic_surface = None
    split_surfaces = list()
    outer_faces_to_remove = list() 
    if (surface2.diagonal_length> surface1.diagonal_length):        # switch so that surface 1 is always longer
        temp = deepcopy(surface1)
        surface1 = deepcopy(surface2)
        surface2 = temp
        
        temp = surface1_indx
        surf1_index = surface2_indx
        surf2_index = temp

    if (surface1.diagonal_length> surface2.diagonal_length): # Surface 1 is longer than surface 2, shrink surface 1
        # Shrink surface 1 but keep the same k value since it's periodic direction is "k"
        surface1_full = deepcopy(surface1)
        surface1 = create_face(blocks[block_indx], surface2.IMIN, surface2.IMAX, 
                                    surface2.JMIN, surface2.JMAX, 
                                    surface1.KMIN, surface1.KMAX)


    angle, rotation_matrix = linear_real_transform(surface1,surface2)   # Do linear real transform to get angle and rotation matrix
    # Check the rotation - rotate the corners to see if this is really equal
    corner1,corner2 = surface2.get_corners()
    corner1_rotated = np.matmul(rotation_matrix,np.array(corner1).transpose()) # y[0], y[1], y[2] are the new x1,y1,z1 rotated about x-axis and they should equal x2,y2,z2 if it's periodic
    corner2_rotated = np.matmul(rotation_matrix,np.array(corner2).transpose())
    corner1_to_match,corner2_to_match = surface1.get_corners()

    corner1_match_error = np.sum(np.square(corner1_rotated-corner1_to_match))
    corner2_match_error = np.sum(np.square(corner2_rotated-corner2_to_match))

    if corner1_match_error<1E-10 and corner2_match_error<1E-10: 
        faces = split_face(surface1,blocks[block_indx],
                            surface1_full.IMIN,surface1_full.JMIN,surface1_full.KMIN,
                            surface1_full.IMAX,surface1_full.JMAX,surface1_full.KMAX)
        
        split_surfaces.extend(faces)              # This list will be appended to the outer_faces
        outer_faces_to_remove.append(surface1_indx) # Schedule these two surfaces to be removed from the outer_faces since they match 
        outer_faces_to_remove.append(surface2_indx)
        
        periodic_surface = {'block1':{'index':block_indx,
            'IMIN':surface1.IMIN,'JMIN':surface1.JMIN,'KMIN':surface1.KMIN, 
            'IMAX':surface1.IMAX,'JMAX':surface1.JMAX,'KMAX':surface1.KMAX, },
            'block2':{'index':block_indx,
            'IMIN':surface2.IMIN,'JMIN':surface2.JMIN,'KMIN':surface2.KMIN, 
            'IMAX':surface2.IMAX,'JMAX':surface2.JMAX,'KMAX':surface2.KMAX, } }

    return periodic_surface,split_surfaces, outer_faces_to_remove, angle, rotation_matrix


from .face import Face, split_face, create_face_from_diagonals
from .block import Block, rotate_block, reduce_blocks
from .connectivity import get_outer_faces, find_matching_blocks, get_face_intersection, connectivity
from .read import read_plot3D
from .write import write_plot3D
from .differencing import find_edges, find_face_edges
from .periodicity import find_periodicity
from .point_match import point_match


'''
    This code computes all the edges of a block or block face 
'''
import numpy as np 
import pandas as pd 

def find_face_edges(X:np.ndarray,Y:np.ndarray,Z:np.ndarray):
    """Check if the edges of both faces to see if they are parallel. Face can be in any direction (I,J) (I,K) etc.
        if edges are parallel then their vertices might intersect
        Find edges will always deal with faces and not something that is in 3D

    Args:
        X (np.ndarray): Multi-dimensional array (2 dimensions PMAX,QMAX ) representing values of X for the block domain
        Y (np.ndarray): Multi-dimensional array (2 dimensions PMAX,QMAX ) representing values of Y for the block domain
        Z (np.ndarray): Multi-dimensional array (2 dimensions PMAX,QMAX ) representing values of Z for the block domain

    Returns:
        pandas.DataFrame: with columns p,q,dp,dq where dp, dq are each tuples containing (dx_b,dy_b,dz_b),(dx_f,dy_f,dz_f)
    """
    (PMAX,QMAX) = X.shape
    diffArray = list()

    for p in range(0,PMAX):
        for q in range(0,QMAX):
                dx_b = 0; dy_b = 0; dz_b = 0    # Preset to 0
                if p!=0:
                    dx_b = X[p-1,q]- X[p,q]
                    dy_b = Y[p-1,q]- Y[p,q]
                    dz_b = Z[p-1,q]- Z[p,q]

                dx_f = 0; dy_f = 0; dz_f = 0    # Preset to 0
                if p!=PMAX-1:
                    dx_f = X[p+1,q]-  X[p,q]
                    dy_f = Y[p+1,q] - Y[p,q]
                    dz_f = Z[p+1,q] - Z[p,q] 

                dp = ((dx_b,dy_b,dz_b),(dx_f,dy_f,dz_f))                
                if q!=0:
                    dx_b = X[p,q-1] - X[p,q]
                    dy_b = Y[p,q-1] - Y[p,q]
                    dz_b = Z[p,q-1] - Z[p,q]

                if q!=QMAX-1:
                    dx_f = X[p,q+1] - X[p,q]
                    dy_f = Y[p,q+1] - Y[p,q]
                    dz_f = Z[p,q+1] - Z[p,q]
                dq = ((dx_b,dy_b,dz_b),(dx_f,dy_f,dz_f))

                diffArray.append({"p":p,"q":q,'dp':dp,'dq':dq})

    df = pd.DataFrame(data = diffArray)
    return df 

def find_edges(X:np.ndarray,Y:np.ndarray,Z:np.ndarray):
    """Check if the edges of both blocks that are parallel. Takes into account the whole block and not a single face
        if edges are parallel then their vertices might intersect. 

    Args:
        X (np.ndarray): Multi-dimensional array (3 dimensions IMAX,JMAX,KMAX ) representing values of X for the block/face domain
        Y (np.ndarray): Multi-dimensional array (3 dimensions IMAX,JMAX,KMAX ) representing values of Y for the block/face domain
        Z (np.ndarray): Multi-dimensional array (3 dimensions IMAX,JMAX,KMAX ) representing values of Z for the block/face domain

    Returns:
        pandas.DataFrame: Dataframe with columns i,j,k,di,dj,dk where di,dj,dk are each tuples containing (dx_b,dy_b,dz_b),(dx_f,dy_f,dz_f)
                _f = forward differencing
    """
    (IMAX,JMAX, KMAX) = X.shape
    diffArray = list()

    for i in range(0,IMAX):
        for j in range(0,JMAX):
            for k in range(0,KMAX):
                dx_b = 0; dy_b = 0; dz_b = 0    # Preset to 0
                if i!=0:
                    # Backward Differencing i
                    dx_b = X[i-1,j,k]- X[i,j,k]
                    dy_b = Y[i-1,j,k]- Y[i,j,k]
                    dz_b = Z[i-1,j,k]- Z[i,j,k]

                dx_f = 0; dy_f = 0; dz_f = 0    # Preset to 0
                if i!=IMAX-1:
                    # Forward Differencing i
                    dx_f = X[i,j,k] - X[i+1,j,k]
                    dy_f = Y[i,j,k] - Y[i+1,j,k]
                    dz_f = Z[i,j,k] - Z[i+1,j,k]

                di = ((dx_b,dy_b,dz_b),(dx_f,dy_f,dz_f))                
                if j!=0:
                    # Backward Differencing j
                    dx_b = X[i,j-1,k]- X[i,j,k]
                    dy_b = Y[i,j-1,k]- Y[i,j,k]
                    dz_b = Z[i,j-1,k]- Z[i,j,k]

                if j!=JMAX-1:
                    # Forward Differencing j
                    dx_f = X[i,j,k] - X[i,j+1,k]
                    dy_f = Y[i,j,k] - Y[i,j+1,k]
                    dz_f = Z[i,j,k] - Z[i,j+1,k]
                dj = ((dx_b,dy_b,dz_b),(dx_f,dy_f,dz_f))

                dx_f = 0; dy_f = 0; dz_f = 0    # Preset to 0
                if k!=0:
                    # Backward Differencing k
                    dx_b = X[i,j,k-1]- X[i,j,k]
                    dy_b = Y[i,j,k-1]- Y[i,j,k]
                    dz_b = Z[i,j,k-1]- Z[i,j,k]

                if k!=KMAX-1:
                    # Forward Differencing k
                    dx_f = X[i,j,k] - X[i,j,k+1]
                    dy_f = Y[i,j,k] - Y[i,j,k+1]
                    dz_f = Z[i,j,k] - Z[i,j,k+1]
                dk = ((dx_b,dy_b,dz_b),(dx_f,dy_f,dz_f))
                
                diffArray.append({"i":i,"j":j,"k":k,'di':di,'dj':dj,'dk':dk})
    df = pd.DataFrame(data =diffArray)
    return df 


from os import write
import numpy as np 
import os.path as osp
import struct
from typing import List

from pandas.core.indexing import need_slice
from .block import Block

def __write_plot3D_block_binary(f,B:Block):
    """Write binary plot3D block which contains X,Y,Z
        default format is Big-Endian

    Args:
        f (IO): file handle
        B (Block): writes a single block to a file
    """
    '''
        https://docs.python.org/3/library/struct.html
    '''
    def write_var(V:np.ndarray):
        for k in range(B.KMAX):
            for j in range(B.JMAX):
                for i in range(B.IMAX):
                    f.write(struct.pack('>f',V[i,j,k]))
    write_var(B.X)
    write_var(B.Y)
    write_var(B.Z)


def __write_plot3D_block_ASCII(f,B:Block,columns:int=6):
    """Write plot3D block in ascii format 

    Args:
        f (IO): file handle
        B (Block): writes a single block to a file
        columns (int, optional): Number of columns in the file. Defaults to 6.
    """
    def write_var(V:np.ndarray):
        bNewLine = False
        indx = 0
        for k in range(B.KMAX):
            for j in range(B.JMAX):
                for i in range(B.IMAX):
                    f.write('{0:8.8f} '.format(V[i,j,k]))
                    bNewLine=False
                    indx+=1
                    if (indx % columns) == 0:
                        f.write('\n')
                        bNewLine=True
                    
        if not bNewLine:
            f.write('\n')
    write_var(B.X)
    write_var(B.Y)
    write_var(B.Z)

def write_plot3D(filename:str,blocks:List[Block],binary:bool=True):
    """Writes blocks to a Plot3D file

    Args:
        filename (str): name of the file to create 
        blocks (List[Block]): List containing all the blocks to write
        binary (bool, optional): Format required. Defaults to True.
    """
    if binary:
        with open(filename,'wb') as f:
            f.write(struct.pack('>I',len(blocks)))
            for b in blocks:
                IMAX,JMAX,KMAX = b.X.shape
                f.write(struct.pack('>I',IMAX))
                f.write(struct.pack('>I',JMAX))
                f.write(struct.pack('>I',KMAX))
            for b in blocks:
                __write_plot3D_block_binary(f,b)
    else:
        with open(filename,'w') as f:
            f.write('{0:d}\n'.format(len(blocks)))
            for b in blocks:
                IMAX,JMAX,KMAX = b.X.shape
                f.write('{0:d} {1:d} {2:d}\n'.format(IMAX,JMAX,KMAX))            
            for b in blocks:
                __write_plot3D_block_ASCII(f,b)


from pandas.core.algorithms import unique
from .block import Block
from .face import Face, create_face_from_diagonals, split_face
import math 
from itertools import product, combinations
from tqdm import trange
import numpy as np 
from .differencing import find_face_edges
import pandas as pd
from operator import eq 
import tqdm
from typing import List
import math
from .point_match import point_match

def get_outer_faces(block1:Block):
    """Get the outer faces of a block

    Args:
        block1 (Block): A plot3D block

    Returns:
        List[Face]: Non matching faces of the block 
        List[(Face,Face)]: Matching faces inside the block 
    """
    I = [0,block1.IMAX-1]               # Python index starts at 0, need to subtract 1 for it to get the i,j,k
    J = [0,block1.JMAX-1]
    K = [0,block1.KMAX-1]
    # Create the outer faces        
    faces = list()
    face = Face(4)
    i=I[0]
    for j in J:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    i=I[1]
    for j in J:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    j=J[0]
    for i in I:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    j=J[1]
    for i in I:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)

    faces.append(face)
    face = Face(4)
    k=K[0]
    for i in I:
        for j in J:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    k=K[1]
    for i in I:
        for j in J:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    faces.append(face)

    # Check if faces match each other
    matching = list()
    non_matching = list()
    for i in range(len(faces)):
        matchFound = False
        for j in range(len(faces)):
            if (i!=j and faces[i] == faces[j]):
                matching.append((i,j))
                matchFound = True
        if not matchFound:
            non_matching.append(faces[i]) # these are guaranteed to be exterior 
    matching = list(unique_pairs(matching))
    matching = [(faces[i],faces[j]) for i,j in matching]
    
    # Make sure normals do not intersect 
    # block_center_to_face_center =  block1.cx
    return non_matching, matching # these should be the outer faces

def unique_pairs(listOfItems:list):
    """Checks if an item is not already in the list 
    
    Args:
        listOfItems (list): list of combinations e.g. (1,2),(3,4),(2,1)

    Yields:
        unique pair: [description]
    """
    seen = set()  #use set to keep track of already seen items, sets provide O(1) lookup  
    for x,y in listOfItems:
        if x!=y and (y,x) not in seen:
            seen.add((x,y)) 
            yield x,y


def find_matching_blocks(block1:Block,block2:Block, full_face_match=False):  
    """Takes two blocks and finds all matching pairs

    Args:
        block1 (Block): Any plot3d Block that is not the same as block2
        block2 (Block): Any plot3d Block that is not the same as block1
        full_face_match (bool): use full face matching (Much faster)

    Returns:
        pandas.DataFrame: corners of matching pair as block1_corners,block2_corners ([imin,jmin,kmin],[imax,jmax,kmax]), ([imin,jmin,kmin],[imax,jmax,kmax])
    """
    # Check to see if outer face of block 1 matches any of the outer faces of block 2
    block_match_indices = list()

    block1_outer,_ = get_outer_faces(block1)
    block2_outer,_ = get_outer_faces(block2)
    block1_split_faces = list()
    block2_split_faces = list() 
    # Create a dataframe for block1 and block 2 inner matches, add to df later
    # df,split_faces1,split_faces2 = get_face_intersection(block1_outer[3],block2_outer[4],block1,block2)                

    # Checks the nodes of the outer faces to see if any of them match 
    block1MatchingFace = list()
    block2MatchingFace = list()
    match = True
    while (match):
        match = False
        for p in range(len(block1_outer)):
            block1_face = block1_outer[p]
            for q in range(len(block2_outer)):
                block2_face = block2_outer[q]
                df,split_faces1,split_faces2 = get_face_intersection(block1_face,block2_face,block1,block2,full_face_match)                
                if len(df)>2:
                    block_match_indices.append(df)
                    block1MatchingFace.append(block1_face)
                    block2MatchingFace.append(block2_face)
                    block1_split_faces.extend(split_faces1)
                    block2_split_faces.extend(split_faces2)
                    match = True

        [block1_outer.remove(b) for b in block1MatchingFace if b in block1_outer]
        [block2_outer.remove(b) for b in block2MatchingFace if b in block2_outer]
        if len(block1_split_faces)>0:
            block1_outer.extend(block1_split_faces)
            block1_split_faces.clear()
        if len(block2_split_faces)>0:
            block2_outer.extend(block2_split_faces)
            block2_split_faces.clear()
    
    return block_match_indices, block1_outer, block2_outer # Remove duplicates using set and list 



def get_face_intersection(face1:Face,face2:Face,block1:Block,block2:Block,full_face_match=False,tol:float=1E-6):
    """Get the index of the intersection between two faces located on two different blocks 

    Args:
        face1 (Face): An exterior face
        face2 (Face): An exterior face from a different block
        block1 (Block): block containing face1
        block2 (Block): block containing face2
        full_face_match (bool): use full face matching (Much faster)

    Returns:
        (Tuple): containing

            - (pandas.DataFrame): dataframe with matches. Columns = I1, J1, K1, I2, J2, K2
            - (List[Face]): any split faces from block 1
            - (List[Face]): any split faces from block 2 
    """

    def select_multi_dimensional(T:np.ndarray,dim1:tuple,dim2:tuple, dim3:tuple):
        """Takes a block (T) and selects X,Y,Z from the block given a face's dimensions
            theres really no good way to do this in python 
        Args:
            T (np.ndarray): arbitrary array so say a full matrix containing X
            dim1 (tuple): 20,50 this selects X in the i direction from i=20 to 50
            dim2 (tuple): 40,60 this selects X in the j direction from j=40 to 60
            dim3 (tuple): 10,20 this selects X in the k direction from k=10 to 20

        Returns:
            np.ndarray: returns X or Y or Z given some range of I,J,K
        """
        if dim1[0] == dim1[1]:
            return T[ dim1[0], dim2[0]:dim2[1]+1, dim3[0]:dim3[1]+1 ]
        if dim2[0] == dim2[1]:
            return T[ dim1[0]:dim1[1]+1, dim2[0], dim3[0]:dim3[1]+1 ]
        if dim3[0] == dim3[1]:
            return T[ dim1[0]:dim1[1]+1, dim2[0]:dim2[1]+1, dim3[0] ]
        
        return T[dim1[0]:dim1[1], dim2[0]:dim2[1], dim3[0]:dim3[1]]
    
    match_location = list()
    df =pd.DataFrame(columns=['i1','j1','k1','i2','j2','k2'])
    split_faces1 = None
    split_faces2 = None
    matchedIndicies = face1.match_indices(face2) # face1 == face2
    if len(matchedIndicies)==4: # Checks to see if the two faces corners are actually equal
        for i,j in matchedIndicies:
            i1,j1,k1 = face1.I[i],face1.J[i],face1.K[i]
            i2,j2,k2 = face2.I[j],face2.J[j],face2.K[j]
            match_location.append({"i1":i1,"j1":j1,"k1":k1,'i2':i2,'j2':j2,'k2':k2})
        df = df.append(match_location,ignore_index=True)
        split_faces1 = list()
        split_faces2 = list() 
    elif(not full_face_match): # Check all points interior of the block
        I1 = [face1.IMIN,face1.IMAX]
        J1 = [face1.JMIN,face1.JMAX]
        K1 = [face1.KMIN,face1.KMAX]

        I2 = [face2.IMIN,face2.IMAX]
        J2 = [face2.JMIN,face2.JMAX]
        K2 = [face2.KMIN,face2.KMAX]

        X1 = select_multi_dimensional(block1.X, (I1[0],I1[1]),(J1[0],J1[1]),(K1[0],K1[1]))
        Y1 = select_multi_dimensional(block1.Y, (I1[0],I1[1]),(J1[0],J1[1]),(K1[0],K1[1]))
        Z1 = select_multi_dimensional(block1.Z, (I1[0],I1[1]),(J1[0],J1[1]),(K1[0],K1[1]))

        X2 = select_multi_dimensional(block2.X, (I2[0],I2[1]),(J2[0],J2[1]),(K2[0],K2[1]))
        Y2 = select_multi_dimensional(block2.Y, (I2[0],I2[1]),(J2[0],J2[1]),(K2[0],K2[1]))
        Z2 = select_multi_dimensional(block2.Z, (I2[0],I2[1]),(J2[0],J2[1]),(K2[0],K2[1]))

        # General Search

        if I1[0] == I1[1]: # I is constant in Face 1
            i = I1[0]
            combo = product(range(J1[0],J1[1]+1), range(K1[0],K1[1]+1))
            for c in combo:
                p, q = c
                x = X1[p,q]
                y = Y1[p,q]
                z = Z1[p,q]
                block2_match_location = point_match(x, y, z, X2, Y2, Z2,tol)
                if sum(block2_match_location)!=-2:
                    p2 = int(block2_match_location[0])
                    q2 = int(block2_match_location[1])
                    # if __edge_match2(df1_edges,df2_edges, p, q, p2, q2):
                    if I2[0]==I2[1]:
                        match_location.append({"i1":i,"j1":p,"k1":q,'i2':I2[0],'j2':p2,'k2':q2})
                    if J2[0]==J2[1]:
                        match_location.append({"i1":i,"j1":p,"k1":q,'i2':p2,'j2':J2[0],'k2':q2})
                    if K2[0]==K2[1]:
                        match_location.append({"i1":i,"j1":p,"k1":q,'i2':p2,'j2':q2,'k2':K2[0]})
            df = df.append(match_location,ignore_index=True)

        elif J1[0] == J1[1]: # J is constant in face 1 
            j = J1[0]
            combo = product(range(I1[0],I1[1]+1),range(K1[0],K1[1]+1))
            for c in combo:
                p, q = c
                x = X1[p,q]
                y = Y1[p,q]
                z = Z1[p,q]
                block2_match_location = point_match(x, y, z, X2, Y2, Z2,tol)
                if sum(block2_match_location)!=-2:
                    p2 = int(block2_match_location[0])
                    q2 = int(block2_match_location[1])
                    # if __edge_match2(df1_edges,df2_edges, p, q, p2, q2):
                    if I2[0]==I2[1]:
                        match_location.append({"i1":p,"j1":J1[0],"k1":q,'i2':I2[0],'j2':p2,'k2':q2})
                    if J2[0]==J2[1]:
                        match_location.append({"i1":p,"j1":J1[0],"k1":q,'i2':p2,'j2':J2[0],'k2':q2})
                    if K2[0]==K2[1]:
                        match_location.append({"i1":p,"j1":J1[0],"k1":q,'i2':p2,'j2':q2,'k2':K2[0]})
                df = df.append(match_location,ignore_index=True)

        elif K1[0] == K1[1]: # K is constant in face 1 
            k = K1[0]
            combo = product(range(I1[0],I1[1]+1), range(J1[0],J1[1]+1))
            for c in combo:
                p, q = c
                x = X1[p,q]
                y = Y1[p,q]
                z = Z1[p,q]
                block2_match_location = point_match(x, y, z, X2, Y2, Z2,tol) # pm,qm are the p and q indicies where match occurs 
                if sum(block2_match_location)!=-2:
                    p2 = int(block2_match_location[0])
                    q2 = int(block2_match_location[1])
                    # if __edge_match2(df1_edges,df2_edges, p, q, p2, q2):
                    if I2[0]==I2[1]:
                        match_location.append({"i1":p,"j1":q,"k1":K1[0],'i2':I2[0],'j2':p2,'k2':q2})
                    if J2[0]==J2[1]:
                        match_location.append({"i1":p,"j1":q,"k1":K1[0],'i2':p2,'j2':J2[0],'k2':q2})
                    if K2[0]==K2[1]:
                        match_location.append({"i1":p,"j1":q,"k1":K1[0],'i2':p2,'j2':q2,'k2':K2[0]})
            df = df.append(match_location,ignore_index=True)
        
        # Checking for split faces 
        if len(df)>0:
            if (len(df)==2 or  __check_edge(df)):
                df = pd.DataFrame()
            else:
                # Filter match increasing - This keeps uniqueness
                if I1[0]==I1[1]:
                    df = __filter_block_increasing(df,'j1')
                    df = __filter_block_increasing(df,'k1')
                elif J1[0]==J1[1]:
                    df = __filter_block_increasing(df,'i1')
                    df = __filter_block_increasing(df,'k1')
                elif K1[0]==K1[1]:
                    df = __filter_block_increasing(df,'i1')
                    df = __filter_block_increasing(df,'j1')

                if I2[0]==I2[1]:
                    df = __filter_block_increasing(df,'j2')
                    df = __filter_block_increasing(df,'k2')
                elif J2[0]==J2[1]:
                    df = __filter_block_increasing(df,'i2')
                    df = __filter_block_increasing(df,'k2')
                elif K2[0]==K2[1]:
                    df = __filter_block_increasing(df,'i2')
                    df = __filter_block_increasing(df,'j2')

                # Check for Split faces
                ## Block 1
                main_face = create_face_from_diagonals(block1,imin=I1[0],imax=I1[1], jmin=J1[0],jmax=J1[1],kmin=K1[0],kmax=K1[1])
                imin, jmin, kmin = df['i1'].min(), df['j1'].min(), df['k1'].min()
                imax, jmax, kmax = df['i1'].max(), df['j1'].max(), df['k1'].max()
                if int(imin==imax) + int(jmin==jmax) + int(kmin==kmax)==1:
                    split_faces1 = split_face(main_face,block1,imin=imin,imax=imax,jmin=jmin,jmax=jmax,kmin=kmin,kmax=kmax)

                ## Block 2
                main_face = create_face_from_diagonals(block2,imin=I2[0],imax=I2[1], jmin=J2[0],jmax=J2[1],kmin=K2[0],kmax=K2[1])
                imin, jmin, kmin = df['i2'].min(), df['j2'].min(), df['k2'].min()
                imax, jmax, kmax = df['i2'].max(), df['j2'].max(), df['k2'].max()
                if int(imin==imax) + int(jmin==jmax) + int(kmin==kmax)==1:
                    split_faces2 = split_face(main_face,block2,imin=imin,imax=imax,jmin=jmin,jmax=jmax,kmin=kmin,kmax=kmax)

    return df, split_faces1, split_faces2



def __filter_block_increasing(df:pd.DataFrame,key1:str):
    """Filters dataframe results of get_face_intersection to make sure both key1 is increasing. 
        When searching through a plot3D we check based on the planes e.g. const i, j, or k 
        values will be removed if they are not

    Args:
        df (pd.DataFrame): DataFrame containing matching points 
        key1 (str): column that you want to be in increasing order 

    Returns:
        pd.DataFrame: sorted dataframe
    """        
    key1_vals = list(df[key1].unique()) # get the unique values 
    key1_vals.sort()
    key1_vals_to_use = list()

    for i in range(len(key1_vals)-1):
        if (key1_vals[i+1] - key1_vals[i])==1: # Remove
            key1_vals_to_use.append(key1_vals[i])
    if (key1_vals[-1] - key1_vals[-2])==1: # Remove
            key1_vals_to_use.append(key1_vals[-1])
    df = df[df[key1].isin(key1_vals_to_use)]        
    return df

def __check_edge(df:pd.DataFrame):
    """Check if the results of get_face_intersection is an edge instead of a face
        if it's an edge then both intersecting blocks are connected by an edge on both blocks 

    Args:
        df (pd.DataFrame): dataframe containing columns i1, j1, k1, i2, j2, k2

    Returns:
        boolean: True = It is an edge, False = not edge 
    """
    
    block = [['i1','j1','k1'],['i2','j2','k2']]
    c=0
    for b in block:        
        for k in b:
            c += int(len(df[k].unique())>1) # example i1 = 1,2,3,4,5
    return c<4 #  If "c" is less than 4 then it's an edge 
    

def connectivity(blocks:List[Block],full_face_match=False):
    """Returns a dictionary outlining the connectivity of the blocks along with any exterior surfaces 

    Args:
        blocks (List[Block]): [description]
        full_face_match (bool): assume the entire face of a block is a perfect match on the opposite block (much faster)

    Returns:
        [list]: All matching faces formatted as a list of { 'block1': {'index', 'IMIN', 'JMIN','KMIN', 'IMAX','JMAX','KMAX'} }
        [list]: All exterior surfaces formatted as a list of { 'block_index', 'surfaces': [{'IMIN', 'JMIN','KMIN', 'IMAX','JMAX','KMAX', 'ID'}] }
    """

    def combinations_of_nearest_blocks(blocks:List[Block],nearest_nblocks:int=12):
        """Returns the indices of the nearest 6 blocks based on their centroid

        Args:
            block (Block): block you are interested in
            blocks (List[Block]): list of all your blocks

        Returns:
            [type]: [description]
        """
        combos = list(combinations(range(len(blocks)),2))
        if (len(blocks))<=48:
            return combos
        
        distances = [None] * len(combos)
        for indx,(i,j) in enumerate(combos):
            dx = blocks[i].cx - blocks[j].cx
            dy = blocks[i].cy - blocks[j].cy
            dz = blocks[i].cz - blocks[j].cz
            distances[indx] = math.sqrt(dx*dx + dy*dy + dz*dz)

        df = pd.DataFrame(combos, columns =['block_i', 'block_j'])
        df['distance'] = distances 

        block_i = list(df.block_i.unique())        
        main_df = df.copy(deep=True)
        block_df = list()
        for i in block_i:
            df2 = main_df[main_df['block_i']==i].sort_values('distance')
            if len(df2)>nearest_nblocks:
                d = df2.iloc[nearest_nblocks-1].distance
                block_df.append(main_df[(main_df['block_i']==i) & (main_df['distance'] <= d)])
            else:
                block_df.append(main_df[(main_df['block_i']==i)])
        df = pd.concat(block_df)
        subset = df[['block_i', 'block_j']]
        return [tuple(x) for x in subset.to_numpy()] 
        


    def find_block_index_in_outer_faces(outer_faces:List[Face], block_indx:int):
        """Looks through an array of faces (outer_faces). Outer faces can include multiple blocks
            This function finds the face of corresponding block index

        Args:
            outer_faces ([type]): List of faces
            block_indx (int): block index

        Returns:
            [type]: [description]
        """
        for i in range(len(outer_faces)):
            if outer_faces[i]['block'] == block_indx:
                return i
        return -1 
        
    def update_outer_faces(prev_outer_faces:list, new_outer_faces:list):
        """Takes a previous set of outer faces and a new set of outer faces. 
            If a face id is present in the "previous" but not in the "new" remove that face 

        Args:
            prev_outer_faces (list): previous set of non matching outer faces of a given block 
            new_outer_faces (list): new set of non-matching outer faces of the same block
        """                
        face_intersection = list(set(prev_outer_faces) & set(new_outer_faces))             # Find the intersection 
        return face_intersection                                     
            

    outer_faces = list()      
    face_matches = list()
    # Find the 6 nearest Blocks and search through all that.     
    # combos = list(combinations(range(len(blocks)),2))
    combos = combinations_of_nearest_blocks(blocks)
    t = trange(len(combos))
    for indx in t:     # block i        
        i,j = combos[indx]
        t.set_description(f"Checking connections block {i} with {j}")
        # Takes 2 blocks, gets the matching faces exterior faces of both blocks 
        df_matches, blocki_outerfaces, blockj_outerfaces = find_matching_blocks(blocks[i],blocks[j],full_face_match)
        # Update connectivity for blocks with matching faces 
        if (len(df_matches)>0):
            for df in df_matches:
                face_matches.append({'block1':{
                                            'index':i,'IMIN':df['i1'].min(),'JMIN':df['j1'].min(),'KMIN':df['k1'].min(),
                                            'IMAX':df['i1'].max(),'JMAX':df['j1'].max(),'KMAX':df['k1'].max()
                                        },
                                    'block2':{
                                            'index':j,'IMIN':df['i2'].min(),'JMIN':df['j2'].min(),'KMIN':df['k2'].min(),
                                            'IMAX':df['i2'].max(),'JMAX':df['j2'].max(),'KMAX':df['k2'].max()
                                        },
                                    'match':df})

        # Update Outer Faces 
        ## 01 Find block i and block j
        block_iloc = find_block_index_in_outer_faces(outer_faces,i)
        block_jloc = find_block_index_in_outer_faces(outer_faces,j)
        
        # Update or Append to list of outer faces 
        if block_iloc==-1:      # outer face does not exist 
            outer_faces.append({"block":i, "outer_faces":blocki_outerfaces})                            
        else:                   # Block i is in outer faces 
            outer_faces[block_iloc]['outer_faces'] = update_outer_faces(outer_faces[block_iloc]['outer_faces'], blocki_outerfaces)
        
        if block_jloc==-1:
            outer_faces.append({"block":j, "outer_faces":blockj_outerfaces})
        else:                   # Block j is in outer faces 
            outer_faces[block_jloc]['outer_faces'] = update_outer_faces(outer_faces[block_jloc]['outer_faces'], blockj_outerfaces)
        
    # Find self-matches, do any faces of for example block1 match another face in block 1
    for i in range(len(blocks)):
        _,self_matches = get_outer_faces(blocks[i]) 
        for match in self_matches: # Append to face matches 
            face_matches.append({'block1':{
                                            'index':i,'IMIN':match[0].I.min(),'JMIN':match[0].J.min(),'KMIN':match[0].K.min(),
                                            'IMAX':match[0].I.max(),'JMAX':match[0].J.max(),'KMAX':match[0].K.max()
                                        },
                                    'block2':{
                                            'index':i,'IMIN':match[1].I.min(),'JMIN':match[1].J.min(),'KMIN':match[1].K.min(),
                                            'IMAX':match[1].I.max(),'JMAX':match[1].J.max(),'KMAX':match[1].K.max()
                                        },
                                    'match':pd.DataFrame([{
                                            'index':i,'IMIN':match[0].I.min(),'JMIN':match[0].J.min(),'KMIN':match[0].K.min(),
                                            'IMAX':match[0].I.max(),'JMAX':match[0].J.max(),'KMAX':match[0].K.max()
                                        },{
                                            'index':i,'IMIN':match[1].I.min(),'JMIN':match[1].J.min(),'KMIN':match[1].K.min(),
                                            'IMAX':match[1].I.max(),'JMAX':match[1].J.max(),'KMAX':match[1].K.max()
                                        }])
                                    })



    # Update the outer faces
    outer_faces_formatted = list() # This will contain 
    id = 1 
    for temp in outer_faces:        
        block = {'index':temp['block']}      # Grabs the index 
        surfaces = list()
        for face in temp['outer_faces']:            
            surfaces.append({ 'IMIN':min(face.I), 'JMIN':min(face.J), 'KMIN':min(face.K),
                                'IMAX':max(face.I), 'JMAX':max(face.J), 'KMAX':max(face.K),
                                'id':id })
            id += 1
        block['surfaces'] = surfaces
        outer_faces_formatted.append(block)

    return face_matches, outer_faces_formatted 



import numpy as np 
import os.path as osp
import struct
from typing import List
from .block import Block

def __read_plot3D_chunk_binary(f,IMAX:int,JMAX:int,KMAX:int, big_endian:bool=False):
    """Reads and formats a binary chunk of data into a plot3D block

    Args:
        f (io): file handle
        IMAX (int): maximum I index
        JMAX (int): maximum J index
        KMAX (int): maximum K index
        big_endian (bool, Optional): Use big endian format for reading binary files. Defaults False.

    Returns:
        numpy.ndarray: Plot3D variable either X,Y, or Z 
    """
    A = np.empty(shape=(IMAX, JMAX, KMAX))
    for k in range(KMAX):
        for j in range(JMAX):
            for i in range(IMAX):
                A[i,j,k] = struct.unpack(">f",f.read(4))[0] if big_endian else struct.unpack("f",f.read(4))[0]
    return A

def __read_plot3D_chunk_ASCII(tokenArray:List[str],offset:int,IMAX:int,JMAX:int,KMAX:int):
    """Reads an ascii chunk of plot3D data into a block 

    Args:
        tokenArray (List[str]): this is a list of strings separated by a space, new line character removed ["12","22", ... etc]
        offset (int): how many entries to skip in the array based on block size (IMAX*JMAX*KMAX) of the previous block
        IMAX (int): maximum I index
        JMAX (int): maximum J index
        KMAX (int): maximum K index 

    Returns:
        numpy.ndarray: Plot3D variable either X,Y, or Z
    """
    '''Works for ASCII files 
    '''
    A = np.empty(shape=(IMAX, JMAX, KMAX))
    for k in range(KMAX):
        for j in range(JMAX):
            for i in range(IMAX):
                A[i,j,k] = tokenArray[offset]
                offset+=1

    return A, offset

def read_plot3D(filename:str, binary:bool=True,big_endian:bool=False):
    """Reads a plot3d file and returns Blocks

    Args:
        filename (str): name of the file to read, .p3d, .xyz, .pdc, .plot3d? 
        binary (bool, optional): indicates if the file is binary. Defaults to True.
        big_endian (bool, optional): use big endian format for reading binary files

    Returns:
        List[Block]: List of blocks insdie the plot3d file
    """
    
    blocks = list()
    if osp.isfile(filename):
        if binary:
            with open(filename,'rb') as f:
                nblocks = struct.unpack(">I",f.read(4))[0] if big_endian else struct.unpack("I",f.read(4))[0] # Read bytes            
                IMAX = list(); JMAX = list(); KMAX = list()
                for b in range(nblocks):
                    if big_endian:
                        IMAX.append(struct.unpack(">I",f.read(4))[0]) # Read bytes
                        JMAX.append(struct.unpack(">I",f.read(4))[0]) # Read bytes
                        KMAX.append(struct.unpack(">I",f.read(4))[0]) # Read bytes
                    else:
                        IMAX.append(struct.unpack("I",f.read(4))[0]) # Read bytes
                        JMAX.append(struct.unpack("I",f.read(4))[0]) # Read bytes
                        KMAX.append(struct.unpack("I",f.read(4))[0]) # Read bytes

                for b in range(nblocks):
                    X = __read_plot3D_chunk_binary(f,IMAX[b],JMAX[b],KMAX[b], big_endian)
                    Y = __read_plot3D_chunk_binary(f,IMAX[b],JMAX[b],KMAX[b], big_endian)
                    Z = __read_plot3D_chunk_binary(f,IMAX[b],JMAX[b],KMAX[b], big_endian)
                    b_temp = Block(X,Y,Z)                    
                    blocks.append(b_temp)
        else:
            with open(filename,'r') as f: 
                nblocks = int(f.readline())
                IMAX = list(); JMAX = list(); KMAX = list()
                
                for b in range(nblocks):
                    tokens = [int(w.replace('\n','')) for w in f.readline().split(' ') if w]
                    IMAX.append(tokens[0])
                    JMAX.append(tokens[1])
                    KMAX.append(tokens[2])
                
                lines = [l.replace('\n','').split(' ') for l in f.readlines()] # Basically an array of strings representing numbers
                lines = [item for sublist in lines for item in sublist]         # Flatten list of lists https://stackabuse.com/python-how-to-flatten-list-of-lists/
 
                tokenArray = [float(entry) for entry in lines if entry] # Convert everything to float 
                offset = 0
                for b in range(nblocks):
                    X, offset = __read_plot3D_chunk_ASCII(tokenArray,offset,IMAX[b],JMAX[b],KMAX[b])
                    Y, offset = __read_plot3D_chunk_ASCII(tokenArray,offset,IMAX[b],JMAX[b],KMAX[b])
                    Z, offset = __read_plot3D_chunk_ASCII(tokenArray,offset,IMAX[b],JMAX[b],KMAX[b])
                    b_temp = Block(X,Y,Z)                    
                    blocks.append(b_temp)
    return blocks





import numpy as np 
import math 
from tqdm import trange
from typing import List

class Block:
    """Plot3D Block definition
    """
    def __init__(self, X:np.ndarray,Y:np.ndarray,Z:np.ndarray):
        """Initializes the block using all the X,Y,Z coordinates of the block

        Args:
            X (np.ndarray): All the X coordinates (i,j,k)
            Y (np.ndarray): All the Y coordinates (i,j,k)
            Z (np.ndarray): All the Z coordinates (i,j,k)

        """
        self.IMAX, self.JMAX, self.KMAX = X.shape
        self.X = X
        self.Y = Y
        self.Z = Z
        # Centroid 
        self.cx = np.mean(X) 
        self.cy = np.mean(Y)
        self.cz = np.mean(Z)
    
    def cylindrical(self):
        """Converts the block to cylindrical coordinate system. The rotation axis is assumed to be "x" direction
        """
        self.r = np.sqrt(self.Z*self.Z + self.Y*self.Y)
        self.theta = np.arctan2(self.Y,self.Z)

    def cell_volumes(self):
        """Compute volume of all cells

        Returns:
            numpy.ndarray: volume of all cells

        Reference:
            Davies, D.E. and Salmond, D.J., "Calculation of the Volume of a General Hexahedron for Flow Predicitons", AIAA Journal, vol. 23, No. 6, pp. 954-956, June 1985.  It is (supposedly) exact for a hexahedral whose faces are bi-linear surfaces (i.e., the simplest surface that can be fit through the four nodes defining the face).
        """
        X = self.X
        Y = self.Y
        Z = self.Z
        a = [np.zeros(shape=(self.IMAX,self.JMAX,self.KMAX))  for _ in range(9)]
        # face csi=const 
        for k in range(1,self.KMAX):
            for j in range(1,self.JMAX):
                for i in range(self.IMAX):          # csi-const
                    dx1 = X[i,j,k-1] - X[i,j-1,k]
                    dy1 = Y[i,j,k-1] - Y[i,j-1,k]
                    dz1 = Z[i,j,k-1] - Z[i,j-1,k]

                    dx2 = X[i,j,k] - X[i,j-1,k-1]
                    dy2 = Y[i,j,k] - Y[i,j-1,k-1]
                    dz2 = Z[i,j,k] - Z[i,j-1,k-1]

                    ax = dy1*dz2 - dz1*dy2
                    ay = dz1*dx2 - dx1*dz2
                    az = dx1*dy2 - dy1*dx2

                    a[0][i,j,k] =  ax*0.5
                    a[1][i,j,k]  = ay*0.5
                    a[2][i,j,k]  = az*0.5

        
        for k in range(1,self.KMAX):
            for j in range(self.JMAX):              # face eta=const 
                for i in range(1,self.IMAX):
                    dx1 = X[i,j,k] - X[i-1,j,k-1]
                    dy1 = Y[i,j,k] - Y[i-1,j,k-1]
                    dz1 = Z[i,j,k] - Z[i-1,j,k-1]

                    dx2 = X[i,j,k-1] - X[i-1,j,k]
                    dy2 = Y[i,j,k-1] - Y[i-1,j,k]
                    dz2 = Z[i,j,k-1] - Z[i-1,j,k]

                    ax = dy1*dz2 - dz1*dy2
                    ay = dz1*dx2 - dx1*dz2
                    az = dx1*dy2 - dy1*dx2

                    a[3][i,j,k] = ax*0.5
                    a[4][i,j,k] = ay*0.5
                    a[5][i,j,k] = az*0.5

        # face zit=const 
        for k in range(self.KMAX):                  # zit=const
            for j in range(1,self.JMAX):            
                for i in range(1,self.IMAX):
                    dx1 = X[i,j,k] - X[i-1,j-1,k]
                    dy1 = Y[i,j,k] - Y[i-1,j-1,k]
                    dz1 = Z[i,j,k] - Z[i-1,j-1,k]

                    dx2 = X[i-1,j,k] - X[i,j-1,k]
                    dy2 = Y[i-1,j,k] - Y[i,j-1,k]
                    dz2 = Z[i-1,j,k] - Z[i,j-1,k]

                    ax = dy1*dz2 - dz1*dy2
                    ay = dz1*dx2 - dx1*dz2
                    az = dx1*dy2 - dy1*dx2

                    a[6][i,j,k] = ax*0.5
                    a[7][i,j,k] = ay*0.5
                    a[8][i,j,k] = az*0.5

        cf = np.zeros(shape=(6,3))
        v = np.zeros(shape=(self.IMAX,self.JMAX,self.KMAX))
        
        for k in trange(1,self.KMAX,desc='Calculating the volumes'):
            for j in range(1,self.JMAX):            
                for i in range(1,self.IMAX):
                    cf[0,0] = X[i-1,j-1,k-1] + X[i-1,j-1,k] + X[i-1,j,k-1] + X[i-1,j,k]
                    cf[0,1] = Y[i-1,j-1,k-1] + Y[i-1,j-1,k] + Y[i-1,j,k-1] + Y[i-1,j,k]
                    cf[0,2] = Z[i-1,j-1,k-1] + Z[i-1,j-1,k] + Z[i-1,j,k-1] + Z[i-1,j,k]
                    cf[1,0] = X[i,j-1,k-1] + X[i,j-1,k] + X[i,j,k-1] + X[i,j,k]
                    cf[1,1] = Y[i,j-1,k-1] + Y[i,j-1,k] + Y[i,j,k-1] + Y[i,j,k]
                    cf[1,2] = Z[i,j-1,k-1] + Z[i,j-1,k] + Z[i,j,k-1] + Z[i,j,k]
                    cf[2,0] = X[i-1,j-1,k-1] + X[i-1,j-1,k] + X[i,j-1,k-1] + X[i,j-1,k]
                    cf[2,1] = Y[i-1,j-1,k-1] + Y[i-1,j-1,k] + Y[i,j-1,k-1] + Y[i,j-1,k]
                    cf[2,2] = Z[i-1,j-1,k-1] + Z[i-1,j-1,k] + Z[i,j-1,k-1] + Z[i,j-1,k]
                    cf[3,0] = X[i-1,j,k-1] + X[i-1,j,k] + X[i,j,k-1] + X[i,j,k]
                    cf[3,1] = Y[i-1,j,k-1] + Y[i-1,j,k] + Y[i,j,k-1] + Y[i,j,k]
                    cf[3,2] = Z[i-1,j,k-1] + Z[i-1,j,k] + Z[i,j,k-1] + Z[i,j,k]
                    cf[4,0] = X[i-1,j-1,k-1] + X[i-1,j,k-1] + X[i,j-1,k-1] + X[i,j,k-1]
                    cf[4,1] = Y[i-1,j-1,k-1] + Y[i-1,j,k-1] + Y[i,j-1,k-1] + Y[i,j,k-1]
                    cf[4,2] = Z[i-1,j-1,k-1] + Z[i-1,j,k-1] + Z[i,j-1,k-1] + Z[i,j,k-1]
                    cf[5,0] = X[i-1,j-1,k] + X[i-1,j,k] + X[i,j-1,k] + X[i,j,k]
                    cf[5,1] = Y[i-1,j-1,k] + Y[i-1,j,k] + Y[i,j-1,k] + Y[i,j,k]
                    cf[5,2] = Z[i-1,j-1,k] + Z[i-1,j,k] + Z[i,j-1,k] + Z[i,j,k]

                    vol12=0
                    for n in range(0,2): # n = 0,1
                        for l in range(0,3): # l = 0,1,2
                            vol12 += math.pow(-1,n+1) *(               
                                        +cf[n,l]*a[l][i-1+n,j,k]
                                        +cf[2+n,l]*a[3+l][i,j-1+n,k]
                                        +cf[4+n,l]*a[6+l][i,j,k-1+n])
                    v[i,j,k]= vol12/12
        return v

    
def rotate_block(block,rotation_matrix:np.ndarray) -> Block:
    """Rotates a block by a rotation matrix 

    Args:
        rotation_matrix (np.ndarray): 3x3 rotation matrix 

    Returns:
        Block: returns a new rotated block 
    """
    X = block.X.copy()
    Y = block.Y.copy()
    Z = block.Z.copy()
    points = np.zeros(shape=(3,block.IMAX*block.JMAX*block.KMAX))
    indx = 0
    for i in range(block.IMAX):
        for j in range(block.JMAX):
            for k in range(block.KMAX):
                points[0,indx] = block.X[i,j,k]
                points[1,indx] = block.Y[i,j,k]
                points[2,indx] = block.Z[i,j,k]
                indx+=1
    points_rotated = np.matmul(rotation_matrix,points)
    indx=0
    for i in range(block.IMAX):
        for j in range(block.JMAX):
            for k in range(block.KMAX):
                X[i,j,k] = points_rotated[0,indx]
                Y[i,j,k] = points_rotated[1,indx]
                Z[i,j,k] = points_rotated[2,indx]
                indx+=1
                
    return Block(X,Y,Z)    
                    


def reduce_blocks(blocks:List[Block],factor:int=2):
    """reduce the blocks by a factor of (factor)

    Args:
        blocks (List[Block]): list of blocks to reduce in size
        factor (int, optional): Number of indicies to skip . Defaults to 2.

    Returns:
        [type]: [description]
    """
    for i in range(len(blocks)):
        blocks[i].X = blocks[i].X[::factor,::factor,::factor]
        blocks[i].Y = blocks[i].Y[::factor,::factor,::factor]
        blocks[i].Z = blocks[i].Z[::factor,::factor,::factor]
        blocks[i].IMAX,blocks[i].JMAX,blocks[i].KMAX = blocks[i].X.shape
    return blocks



from setuptools import setup

_config = {
    "name": "Plot3D",
    "version":"0.1.4",
    "url": "",
    "author": "Paht Juangphanich",
    "author_email": "paht.juangphanich@nasa.gov",    
    "install_requires":['bz2file','tqdm','scipy','pandas','numpy','matplotlib'],
    "packages":["plot3d"],
    'license':"GNU GPLv3",
    'zip_safe':False,
}

def main() -> int:
    """ Execute the setup command.
    """

    def version():
        """ Get the local package version. """
        return _config["version"]

    _config.update({
        "version": version(),
    })

    setup(**_config)
    return 0


# Make the script executable.
if __name__ == "__main__":
    raise SystemExit(main())



import itertools
from typing import List, Tuple
import numpy as np
from numpy.lib import math
from .block import Block

class Face:
    """Defines a Face of a block for example IMIN,JMIN,JMIN to IMAX,JMIN,JMIN
    """
    def __init__(self,nvertex:int=4):
        """ Defines a face using nvertex 4 = quad 3 = triangles 

        Args:
            nvertex (int, optional): Number of vertices. Defaults to 4.
            id (int, optional): A unique index indentifying a face
        """
        self.x = np.zeros(4)
        self.y = np.zeros(4)
        self.z = np.zeros(4)        
        self.I = np.zeros(4,dtype=np.int64)
        self.J = np.zeros(4,dtype=np.int64)
        self.K = np.zeros(4,dtype=np.int64)
        self.cx = 0         # centroid 
        self.cy = 0
        self.cz = 0
        self.nvertex=0
    
    @property
    def IMIN(self):
        return self.I.min()
    
    @property
    def JMIN(self):
        return self.J.min()
    
    @property
    def KMIN(self):
        return self.K.min()
    
    @property
    def IMAX(self):
        return self.I.max()
    
    @property
    def JMAX(self):
        return self.J.max()
    
    @property
    def KMAX(self):
        return self.K.max()
    
    @property
    def isEdge(self):
        """check if the face is actually an edge. This is an edge if two indicies IMIN == IMAX or JMIN=JMAX or KMIN=KMAX

        Returns:
            [bool]: True if face is really an edge 
        """
        return (int(self.IMIN == self.IMAX) + int(self.JMIN == self.JMAX) + int(self.KMIN == self.KMAX)) > 1

    @property
    def isPoint(self):
        """check if the face is actually an edge. This is an edge if two indicies IMIN == IMAX or JMIN=JMAX or KMIN=KMAX

        Returns:
            [type]: True if face is really a point 
        """
        return (int(self.IMIN == self.IMAX) + int(self.JMIN == self.JMAX) + int(self.KMIN == self.KMAX)) > 2

    @property
    def get_val(self,i_val:int,j_val:int,k_val:int):
        """Get the value where key (I,J,K) is equal to val

        Args:            
            i_val (int): value of I
            j_val (int): value of J
            k_val (int): value of K

        Returns:
            [float]: x value
            [float]: y value
            [float]: z value
        """
        
        indx_i = np.where(self.I == i_val).tolist()
        indx_j = np.where(self.J == j_val).tolist()
        indx_k = np.where(self.K == k_val).tolist()
        
        indx_i.extend(indx_j)
        indx_i.extend(indx_k)

        indx = list(set([indx_i]))[0] # Get the common one through a union
        return self.x[indx], self.y[indx], self.z[indx]


    def add_vertex(self, x:float,y:float,z:float, i:int, j:int, k:int):
        """Add vertex to define a face

        Args:
            x (float): x-coordinate
            y (float): y-coordinate
            z (float): z-coordinate
            i (int): i-th index of the coordinates (x,y,z)
            j (int): j-th index of the coordinates (x,y,z)
            k (int): k-th index of the coordinates (x,y,z)
        """
        
        self.x[self.nvertex] = x
        self.y[self.nvertex] = y 
        self.z[self.nvertex] = z 
        self.I[self.nvertex] = i
        self.J[self.nvertex] = j
        self.K[self.nvertex] = k        
        self.nvertex+=1    
        if self.nvertex==4:
            self.cx = self.x.mean()
            self.cy = self.y.mean()
            self.cz = self.z.mean()

    def __normal__(self):
        """Computes the normal vector of the face 
            not really used but if anyone wants it. 
        """
        if (self.I[0]!=self.I[1]) and (self.I[0]!=self.I[2]):
            indx = np.argsort(self.I)
        elif (self.J[0]!=self.J[1]) and (self.J[0]!=self.J[2]):
            indx = np.argsort(self.J)
        elif (self.K[0]!=self.K[1]) and (self.K[0]!=self.K[2]):
            indx = np.argsort(self.K)

        self.x = self.x[indx]
        self.y = self.y[indx]
        self.z = self.z[indx]
        self.I = self.I[indx]
        self.J = self.J[indx]
        self.K = self.K[indx]
        x1 = self.x[1]-self.x[0]; y1 = self.y[1]-self.y[0]; z1 = self.z[1]-self.z[0]
        x2 = self.x[2]-self.x[0]; y2 = self.y[2]-self.y[0]; z2 = self.z[2]-self.z[0]
        nx = y1*z2-y2*z1; ny = -1*(x1*z2-x2*z1); nz = x1*y2-x2*y1
        self.nx = nx
        self.ny = ny
        self.nz = nz 

    def match_indices(self,f):
        """Check to see if two faces are the same. Checks to see if any of vertices x,y,z match
            Normally this is used by Face1==Face2

        Args:
            f (Face): another face

        Returns:
            List[(int,int)]: list of indicies where there's a match. 
        """
        matched_vertices = list()
        tol = 1E-6
        matchedIndices = list()
        for i in range(self.nvertex):
            for j in range(f.nvertex):
                dx = abs(self.x[i] - f.x[j])
                dy = abs(self.y[i] - f.y[j])
                dz = abs(self.z[i] - f.z[j])
                if (dx<tol and dy<tol and dz<tol and (j not in matched_vertices)):
                    matchedIndices.append([i,j])
                    matched_vertices.append(j) # This vertex has been matched, remove from list
                    break # each index can only have a single match
        return matchedIndices

    def __eq__(self, f):
        """Check to see if two faces are the same by looking at the value of their vertices 
        Checks to see if any of vertices x,y,z match

        Args:
            f (Face): another face 

        Returns:
            Boolean: True if faces match, False if no match is found 
        """
        matchedIndices = self.match_indices(f)
        return (len(matchedIndices)==self.nvertex)
        
    def __ne__(self,f):
        """Checks if two faces are not equal 

        Args:
            f (Face): another face 

        Returns:
            Boolean: True if faces match, False if no match is found 
        """
        match = self.__eq__(f)
        return not match
    
    def index_equals(self,f2):
        """Check to see of the face indices are equal

        Args:
            f2 ([type]): [description]
        """
        if (self.IMIN == f2.IMIN and 
            self.JMIN == f2.JMIN and 
            self.KMIN == f2.KMIN and 
            self.IMAX == f2.IMAX and 
            self.JMAX == f2.JMAX and 
            self.KMAX == f2.KMAX):
            return True
    def __hash__(self):
        if (len(self.I)>0):
            return hash((self.I[0], self.J[0], self.K[0], self.I[-1], self.J[-1], self.K[-1]))
        else:
            return hash((0, 0, 0, 0, 0, 0))
            
    def __str__(self):
        if (len(self.I)>0):
            return '{:d},{:d},{:d},{:d},{:d},{:d}'.format(self.IMIN, self.JMIN, self.KMIN, self.IMAX, self.JMAX, self.KMAX)
        else:
            return '{:d},{:d},{:d},{:d},{:d},{:d}'.format(0,0,0,0,0,0)
    
    def __repr__(self):
        return str(self)
    
    @property
    def diagonal_length(self) -> float:
        """Returns the diagonal length of the face 

        Returns:
            float: diagonal length computed using IMIN, IMAX, JMIN, JMAX, KMIN, KMAX 
        """
        minIndx = 0; maxIndx = 0 
        for indx in range(len(self.I)):
            if self.I[indx] == self.IMIN and self.J[indx] == self.JMIN and self.K[indx] == self.KMIN:
                minIndx = indx
            if self.I[indx] == self.IMAX and self.J[indx] == self.JMAX and self.K[indx] == self.KMAX:
                maxIndx = indx
        dx = self.x[minIndx] - self.x[maxIndx]
        dy = self.y[minIndx] - self.y[maxIndx]
        dz = self.z[minIndx] - self.z[maxIndx]
        return math.sqrt(dx*dx + dy*dy + dz*dz)

    def get_corners(self) -> Tuple:
        """Get the corners defined by (IMIN,JMIN,KMIN), (IMAX,JMAX,KMAX),
        
        Returns:
            Tuple: containing

                - **(x,y,z)** (float,float,float): at IMIN,JMIN,KMIN
                - **(x,y,z)** (float,float,float): at IMAX,JMAX,KMAX
        
        Reference: 
            - GlennHT source code https://gitlab.grc.nasa.gov/lte-turbo/GlennHT/-/blob/master/src/M_ccMBMesh.F function computeLRT

        """
        minIndx = 0; maxIndx = 0 
        for indx in range(len(self.I)):
            if self.I[indx] == self.IMIN and self.J[indx] == self.JMIN and self.K[indx] == self.KMIN:
                minIndx = indx
            if self.I[indx] == self.IMAX and self.J[indx] == self.JMAX and self.K[indx] == self.KMAX:
                maxIndx = indx
        return (self.x[minIndx],self.y[minIndx], self.z[minIndx]),(self.x[maxIndx],self.y[maxIndx], self.z[maxIndx])
        


def create_face_from_diagonals(block:Block,imin:int,jmin:int,kmin:int,imax:int,jmax:int,kmax:int):
    """Creates a face on a block given a the diagonals defined as (IMIN,JMIN,KMIN), (IMAX, JMAX, KMAX)

    Args:
        block (Block): Block to create a face on 
        imin (int): Lower Corner IMIN
        jmin (int): Lower Corner JMIN
        kmin (int): Lower Corner KMIN
        imax (int): Upper Corner IMAX
        jmax (int): Upper Corner JMAX
        kmax (int): Upper Corner

    Returns:
        [type]: [description]
    """
    newFace = Face(4)           # This is because two of the corners either imin or imax can be equal
    if imin==imax:
        i = imin
        for j in [jmin,jmax]:
            for k in [kmin,kmax]:
                x = block.X[i,j,k]
                y = block.Y[i,j,k]
                z = block.Z[i,j,k]
                newFace.add_vertex(x,y,z,i,j,k)
    elif jmin==jmax:
        j = jmin
        for i in [imin,imax]:
            for k in [kmin,kmax]:
                x = block.X[i,j,k]
                y = block.Y[i,j,k]
                z = block.Z[i,j,k]
                newFace.add_vertex(x,y,z,i,j,k)
    elif kmin==kmax:
        k = kmin
        for i in [imin,imax]:
            for j in [jmin,jmax]:
                x = block.X[i,j,k]
                y = block.Y[i,j,k]
                z = block.Z[i,j,k]
                newFace.add_vertex(x,y,z,i,j,k)
    return newFace


def split_face(face_to_split:Face, block:Block,imin:int,jmin:int,kmin:int,imax:int,jmax:int,kmax:int):
    """Splits a face with another face within the same block 
        picture the split as a two rectangles inside each other

    Args:
        face_to_split (Face): Face on the block to be split 
        block (Block): Block the split is occuring on 
        imin (int): IMIN index of the split (diagonals)
        jmin (int): JMIN index of the split (diagonals)
        kmin (int): KMIN index of the split (diagonals) 
        imax (int): IMAX index of the split
        jmax (int): JMAX index of the split
        kmax (int): KMAX index of the split 

    :: 

                    left face    top face     right face
         ________       __          __            __
        |   __   |     |  |        |__|          |  |     __
        |  |__|  |     |  |         __           |  |    |__|  face_to_split/center face 
        |________|     |__|        |__|          |__|
                                bottom face

    Returns:
        [List[Faces]]: List of unique faces from the split 
    """
    center_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=kmin,kmax=kmax)

    if kmin == kmax:
        # In the picture above Horizontal = i, vertical = j
        left_face = create_face_from_diagonals(block,
                imin=face_to_split.IMIN,imax=imin,
                jmin=face_to_split.JMIN,jmax=face_to_split.JMAX,
                kmin=kmin, kmax=kmax)

        
        right_face = create_face_from_diagonals(block,
                imin=imax, imax=face_to_split.IMAX,
                jmin=face_to_split.JMIN, jmax=face_to_split.JMAX,
                kmin=kmin, kmax=kmax)

        top_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmax,jmax=face_to_split.JMAX,
            kmin=kmin,kmax=kmax)
        
        bottom_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=face_to_split.JMIN,jmax=jmin,
            kmin=kmin,kmax=kmax)  

    elif (imin==imax):
        # In the picture above Horizontal = j, vertical = k
        left_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=face_to_split.JMIN, jmax=jmin,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)

        right_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmax, jmax=face_to_split.JMAX,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)

        top_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=kmax,kmax=face_to_split.KMAX)

        bottom_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=face_to_split.KMIN,kmax=kmin)

    elif (jmin==jmax):
        # In the picture above Horizontal = i, vertical = k 
        left_face = create_face_from_diagonals(block,
            imin=face_to_split.IMIN,imax=imin,
            jmin=jmin,jmax=jmax,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)

        right_face = create_face_from_diagonals(block,
            imin=imax,imax=face_to_split.IMAX,
            jmin=jmin,jmax=jmax,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)
        
        top_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=kmax,kmax=face_to_split.KMAX)
        
        bottom_face = create_face_from_diagonals(block,
            imin=imin, imax=imax,
            jmin=jmin, jmax=jmax,
            kmin=face_to_split.KMIN, kmax=kmin)
    
    faces = [top_face,bottom_face,left_face,right_face]
    faces = [f for f in faces if not f.isEdge and not f.index_equals(center_face)] # Remove edges 
    return faces 